//! @file single-torrix-pinv.c
//! @author J. Marcel van der Veer
//!
//! @section Copyright
//!
//! This file is part of Algol68G - an Algol 68 compiler-interpreter.
//! Copyright 2001-2023 J. Marcel van der Veer <algol68g@xs4all.nl>.
//!
//! @section License
//!
//! This program is free software; you can redistribute it and/or modify it 
//! under the terms of the GNU General Public License as published by the 
//! Free Software Foundation; either version 3 of the License, or 
//! (at your option) any later version.
//!
//! This program is distributed in the hope that it will be useful, but 
//! WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
//! or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
//! more details. You should have received a copy of the GNU General Public 
//! License along with this program. If not, see <http://www.gnu.org/licenses/>.

//! @section Synopsis
//!
//! REAL Moore-Penrose pseudo-inverse of a matrix using Golub-Reinsch SVD.

#include "a68g.h"
#include "a68g-torrix.h"

#if defined (HAVE_GSL)

//! @brief PROC pseudo inv = ([, ] REAL, REAL) [, ] REAL

void genie_matrix_pinv_lim (NODE_T * p)
{
// Compute the Moore-Penrose pseudo inverse of a MxN matrix.
// G. Strang. "Linear Algebra and its applications", 2nd edition.
// Academic Press [1980].
  gsl_error_handler_t *save_handler = gsl_set_error_handler (torrix_error_handler);
  torrix_error_node = p;
  A68_REAL rcond;
  POP_OBJECT (p, &rcond, A68_REAL);
  gsl_matrix *Z = pop_matrix (p, A68_TRUE), *A;
  unt M = SIZE1 (Z), N = SIZE2 (Z);  
// GSL only handles M>=N, transpose commutes with pseudo inverse.    
  BOOL_T transpose = (M < N);
  if (transpose) {
    A = gsl_matrix_calloc (N, M);
    gsl_matrix_transpose_memcpy (A, Z);
    M = SIZE1 (A); N = SIZE2 (A);  
  } else {
    A = gsl_matrix_calloc (M, N);
    gsl_matrix_memcpy (A, Z);
  }
  gsl_matrix_free (Z);
// A = USVᵀ by Jacobi, more precise than Golub-Reinsch.
// Note that U is not transposed by the gsl routine!
  gsl_matrix *V = gsl_matrix_calloc (N, N);
  gsl_vector *s = gsl_vector_calloc (N);
  int rc = gsl_linalg_SV_decomp_jacobi (A, V, s);
  torrix_test_error (rc);
// Compute S⁻¹ by taking the reciprocal of each nonzero element on the diagonal. 
// Diagonal elements < 'rcond' * largest singular value, are considered zero.
// This because very small eigenvalues wreak havoc on a pseudo inverse.
// Python NumPy default for rcond is 1e-15, but assumes a Hermitian matrix.
// The default here is more conservative.
  REAL_T lwb = gsl_vector_max (s) * (VALUE (&rcond) > 0 ? VALUE (&rcond) : 1e-9);
  gsl_matrix *S_inv = gsl_matrix_calloc (N, M); // calloc sets matrix to 0.
  for (int i = 0; i < N; i++) {
    REAL_T x = gsl_vector_get (s, i);
    if (x > lwb) {
      gsl_matrix_set (S_inv, i, i, 1 / x);
    } else {
      gsl_matrix_set (S_inv, i, i, 0);
    }
  }
  gsl_vector_free (s);
// GSL SVD yields 'thin' SVD - pad to full matrix by adding zeros.
  gsl_matrix *U = gsl_matrix_calloc (M, M);
  for (int i = 0; i < M; i++) {
    for (int j = 0; j < N; j++) {
      gsl_matrix_set (U, i, j, gsl_matrix_get (A, i, j));
    }
  }
// Compute pseudo inverse A⁻¹ = VS⁻¹Uᵀ. 
  gsl_matrix *P_inv = gsl_matrix_calloc (N, M);
  gsl_matrix *A_inv = gsl_matrix_calloc (N, M);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1, V, S_inv, 0, P_inv);
  gsl_blas_dgemm (CblasNoTrans, CblasTrans, 1, P_inv, U, 0, A_inv);
  gsl_matrix_free (P_inv);
  gsl_matrix_free (S_inv);
  gsl_matrix_free (U);
  gsl_matrix_free (V);
// Push result.
  if (transpose) {
    gsl_matrix *T = gsl_matrix_calloc (M, N);
    gsl_matrix_transpose_memcpy (T, A_inv);
    push_matrix (p, T);
    gsl_matrix_free (T);
  } else {
    push_matrix (p, A_inv);
  }
// Clean up.
  gsl_matrix_free (A_inv);
  (void) gsl_set_error_handler (save_handler);
}

//! @brief OP PINV = ([, ] REAL) [, ] REAL

void genie_matrix_pinv (NODE_T * p)
{
  PUSH_VALUE (p, 0, A68_REAL);
  genie_matrix_pinv_lim (p);
}

#endif
